/* $Id: DRAWxtl3.c 613 2007-01-12 17:29:43Z larry $  */

/* ===================================================================== */
/*     DRAWxtl                                                           */
/*                                                                       */
/* Routine to read crystallographic information such as unit cell, space */
/* group, atom positions, and thermal ellipsoid parameters to generate   */
/* POLYHEDRON, SPHERE, ELLIPSOID and BOND geometry POV and VRML scenes.  */
/*                                                                       */
/* Part 3                                                                */
/*                                                                       */
/* Coded by:   L.W. Finger                                               */
/*            Geophysical Laboratory                                     */
/*            Carnegie Institution of Washington                         */
/*            Washington DC 20015                                        */
/*            Larry.Finger@lwfinger.net                                  */
/*                                                                       */
/* Original Version: April 92                                            */
/*            Revised January 1993 to add bonds.                         */
/*            Converted to AVS Module February 1993.                     */
/*            Thermal Ellipsoids added April 1996                        */
/*            Thermal ellipsoids for POV and VRML finished Nov. 1996     */
/*                                                                       */
/*                                                                       */
/* Converted to standalone POV scene file generator Oct. 1994/March 1995 */
/* and VRML generator added September 1995/March 1996                    */
/* by M. Kroeker, Organic Chemistry, Technische Hochschule Darmstadt,    */
/* Petersenstr. 22, D-64287 Darmstadt, Germany                           */
/*                                                                       */
/* Two versions merged  July 1995                                        */
/* ===================================================================== */

/************************************************************************
 *  This file contains routines:
 * build_poly_list - create list of points for polyhedron generation
 * crystal_compute - do the computations for a crystal
 * draw_cell - draw the unit-cell edges
 * generate_lsq_planes - generate best fitting planes 
 * generate_spheres - generate the spheres in the drawing
 * import_schakal - read structural parameters from a Schakal file
 * Transform_POV_Color - convert color input to POV format
 * Transform_VRML_Color - convert color input to VRML format
 * trim_string - remove any trailing \r, \n or ' ' from the end of a string
 *
 *************************************************************************/

#include "drawxtl.h"
#include "draw_ext.h"
#ifndef WIN32
  #include <unistd.h>
#endif

/* ************************************************************** */
/* ************************************************************** */

void build_poly_list (int numb_list, int polyno, int i0, int i1, float cx[3])

/* routine to take list of 'numb_list' indices in 's_vert'
   and find each plane on surface of convex polyhedron.  It then builds
   a list of pointers to these vertices for drawing program.  The algorithm
   is as follows:
   
   1.  If there are 4 vertices, the polyhedron is a tetrahedron and all
       combinations of 3 vertices constitute a face.

   2.  For more than 4 vertices, some planes will be interior.  These are
       found by choosing three points (indicated by i,j,k) and forming two
       vectors defined by V1 = P(j) - P(i) and V2 = P(k) - P(i).  For each
       additional point (l), a third vector V3 = P(l) - P(i) is formed and
       the volume enclosed by these three vectors V = V3 . V1 x V2 is
       calculated.  If this volume is zero, point l lies in the plane.  If
       the plane is on the surface of the polyhedron, all non-zero volumes
       must have the same sign, which may be positive or negative.
*/
/*  numb_list - number of candidate vertices */
/*  polyno - number of polygon - for polygon color */
/*  i0 - pointer to central atom */
/*  i1 - pointer to an atom in the polyhedron */
/*  cx - orthogonal coordinates of center of polyhedron */

{
  int i, j, k, l, m, n;                            /* local loop counters */
  float v0[3], v1[3], v2[3], v3[3], v4[3];      /* vectors used in calculation */
  float p[3];                     /* coordinates of point i */
  float vol, sign_z;
  int ind;                     /* indicator of surface plane */
  int *no;                     /* place to keep vertex id's */
  int imk, jmk, kmk;
  int ii, jj, ll;
  int jplane;
  double phi, cosphi;
  float xm, ym, zm, vx, vy, vz, bm, phi0, vxs, vys, vzs, bms;
  float d1, d2, d3, radius;
  int *ns;
  char edgecolor[40];
  static int warn = 0;
#ifdef OPENGL
  float glr=0.0f, glg=0.0f, glb=0.0f;
#endif
  float Vol = 0.0f;             /* volume of polyhedron */
  float Area, Z0, Z;
  int ih[3];
  float sumth = 0.0f, sumth2 = 0.0f, sumd2 = 0.0f, sumd = 0.0f;
  int nangle = 0;
  int nm, mn;
  float va[3], vb[3], temp, c1, c2=0.0f, vlo, qe=0.0f, siga=0.0f, sigd=0.0f;
  char poly_col_p[40];
  char poly_col_v[40];

  if (!(no = zalloc ((unsigned) ((4 * numb_list) * sizeof (int))))) {
    Error_Box("Unable to get no allocation!");
    return;
  }
  if (!(ns = zalloc ((unsigned) ((4 * numb_list) * sizeof (int))))) {
    Error_Box("Unable to get ns allocation!");
    free (no);
    return;
  }
  radius = rad_edge * 0.005f *Scale;
  memset(no,0,sizeof(no));
  memset(ns,0,sizeof(ns));
  strcpy(poly_col_p,drvui->poly_col[polyno]);
  Transform_POV_Color(poly_col_p);
  strcpy(poly_col_v,drvui->poly_col[polyno]);
  Transform_VRML_Color(poly_col_v);

  if (drvui->p_rad_edge[polyno] != 0.)
    radius = drvui->p_rad_edge[polyno] * 0.005f *Scale;

  Vol = 0.0f;
  for (i=0; i<numb_list-2; i++) {
      ih[0] = i;
      for (j=i+1; j<numb_list-1; j++) {
          ih[1] = j;
          for (k=j+1; k<numb_list; k++) {
              ih[2] = k;
              for (m=0; m<3; m++) {
                  v1[m] = s_vert[3*vertex_list[j]+m] - s_vert[3*vertex_list[i]+m];
                  v2[m] = s_vert[3*vertex_list[k]+m] - s_vert[3*vertex_list[i]+m];
                  v4[m] = s_vert[3*vertex_list[i]+m] - cx[m];
              }
              vcross(v1,v2,v3);
              Area = 0.5f * vlength(v3);
              Z0 = 0.5f * vdot(v3,v4)/Area;
              if (fabs(Z0) < 0.0001) continue;     /* 3 points are in a line */
              for (l=0; l<numb_list; l++) {
                  if (l == i || l == j || l == k) continue;
                  for (m=0; m<3; m++) {
                      v0[m] = s_vert[3*vertex_list[i]+m] - s_vert[3*vertex_list[l]+m];
                  }
                  Z = 0.5f * vdot(v0, v3);
                  if (Z * Z0 < 0.0f) continue;      /* not all points on same side */
                  for (m=0; m<3; m++) {
                      v3[m] *= 0.5f/Area;           /* directions cosines of plane normal */
                  }
              }
              Vol += Area * (float)(fabs(Z0)/3.0);
              for (l=0; l<2; l++) {
                  nm = ih[l];
                  for (m=l+1; m<3; m++) {
                      mn = ih[m];
                      for (n=0; n<3; n++) {
                          va[n] = s_vert[3*vertex_list[nm]+n] - cx[n];
                          vb[n] = s_vert[3*vertex_list[mn]+n] - cx[n];
                      }
                      temp = vdot(va, vb)/(vlength(va) * vlength(vb));
                      temp = min(temp, 1.0f);
                      temp = max(-1.0f,temp);
                      temp = (float)(atan2(sqrt(1.0f - temp*temp),temp) * 180.0/PI);
                      sumth += temp;
                      sumth2 += temp*temp;
                      nangle++;
                  }
              }
          }
      }
  }
  for (i=0; i<numb_list; i++) {
      for (n=0; n<3; n++) {
          va[n] = s_vert[3*vertex_list[i]+n] - cx[n];
      }
      temp = vlength(va);
      sumd2 += temp * temp;
      sumd += temp;

  }
  if (numb_list == 4) {
      c1 = 9.0f * (float)(sqrt(3.0)/8.0);
      c2 = 109.045f;
  } else if (numb_list == 6) {
      c1 = 0.75f;
      c2 = 90.0f;
  } else {
      c1 = 0.0f;
  }
  if (c1) {
      vlo = (float)(exp(2.0 * log(c1 * Vol)/3.0));
      qe = sumd2 / ((float)numb_list * vlo);
      nangle = (nangle + 1)/2;
      sumth2 *= 0.5f;
      sumth *= 0.5f;
      siga = (sumth2 - 2.0f * c2 * sumth + nangle * c2 * c2)/(float)(nangle - 1);
      sigd = (sumd2 - sumd * sumd / (float)numb_list)/(float)(nangle - 1);
  }
  for (i = 0; i < numb_list - 2; ++i) {
    no[0] = vertex_list[i];             /* first point */
    if (DepthCue != 0.) {
/*calculate z coordinate after rotation, and scale edge width accordingly*/
      radius = 0.0f;
      for (l = 0; l <= 2; l++){
        radius += (float) G_Rot[l][2] * s_vert[3 * no[0] + 2];
      }
      radius = rad_edge * 0.005f * Scale + DepthCue * radius;
    }
    for (m = 0; m <= 2; ++m)
      p[m] = s_vert[3 * no[0] + m];
    for (j = i + 1; j < numb_list - 1; ++j) {
      no[1] = vertex_list[j];             /* second point */
      for (k = 0; k <= 2; ++k)
        v1[k] = s_vert[3 * no[1] + k] - p[k];
      for (k = j + 1; k < numb_list; ++k) {
        if (numb_list == 4) {             /* tetrahedron */
          poly_list[draw_list++] = 3;         /* number of vertices for triangle */
          poly_list[draw_list++] = vertex_list[i] + 1;
          poly_list[draw_list++] = vertex_list[j] + 1;
          poly_list[draw_list++] = vertex_list[k] + 1;
          if (draw_list > 4 * NvertM) {
            poly_list = zalloc ((unsigned) (4 * (nvert+2) * sizeof (int)));
            if (!warn) {
              fprintf (drvui->flout,"Overrun of poly_list.\n");
              fprintf (drvui->flout, "Overrun of poly_list.\n");
              Error_Box("Overrun of poly_list. Please send 'str' file\n"
                   "to Larry.Finger@@lwfinger.net.");
              warn = 1;
            }
          }

          imk = vertex_list[i];
          jmk = vertex_list[j];
          kmk = vertex_list[k];
          for (m=0; m<3; m++) {
              v1[m] = s_vert[3*imk+m] - s_vert[3*jmk+m];
              v2[m] = s_vert[3*imk+m] - s_vert[3*kmk+m];
              v3[m] = s_vert[3*kmk+m] - s_vert[3*jmk+m];
          }
          d1 = vlength(v1);
          d2 = vlength(v2);
          d3 = vlength(v3);
          if ((d1 > 0.01) && (d2 > 0.01) && (d3 > 0.01)) {
            fprintf (drvui->fpoutp,
                 " triangle{< %8.5f, %8.5f, %8.5f>,\n",
                 s_vert[3 * imk], s_vert[3 * imk + 1],
                 s_vert[3 * imk + 2]);
            fprintf (drvui->fpoutp,
                 "<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>  texture{pigment{color %s  }}\n",
                 s_vert[3 * jmk], s_vert[3 * jmk + 1],
                 s_vert[3 * jmk + 2], s_vert[3 * kmk],
                 s_vert[3 * kmk + 1], s_vert[3 * kmk + 2],
                 poly_col_p);
            fprintf(drvui->fpoutp," }\n");
#ifdef OPENGL
            if (gl_view) {
                glDisable(GL_LIGHTING);
                glPushMatrix ();
                glLoadName (i0*100000+i0);
                glPushName (i1);
                (void)sscanf (poly_col_v, "%f %f %f", &glr, &glg, &glb);
                glColor3f (glr, glg, glb);
                glBegin (GL_TRIANGLES);
                glVertex3f (s_vert[3 * imk], s_vert[3 * imk + 1], s_vert[3 * imk + 2]);
                glVertex3f (s_vert[3 * jmk], s_vert[3 * jmk + 1], s_vert[3 * jmk + 2]);
                glVertex3f (s_vert[3 * kmk], s_vert[3 * kmk + 1], s_vert[3 * kmk + 2]);
                glEnd ();
                glPopName();
                glPopMatrix ();
                glEnable(GL_LIGHTING);
            }
#endif
            if (edges) {
              float df[3], at[3];
              int m;
              char color[128];
              if (drvui->p_rad_edge[polyno] > 0.005f)
                strncpy (edgecolor, drvui->p_col_edge[polyno],39);
              else
                strncpy (edgecolor, drvui->col_edge,39);
              trim_string(edgecolor,40);
              strcpy(color,edgecolor);
              Transform_POV_Color(edgecolor);
              Transform_VRML_Color(color);
              fprintf (drvui->fpoutp,
                   " cylinder { <%f,%f,%f>,<%f,%f,%f>,%f\n",
                   s_vert[3 * imk], s_vert[3 * imk + 1],
                   s_vert[3 * imk + 2], s_vert[3 * jmk],
                   s_vert[3 * jmk + 1], s_vert[3 * jmk + 2], radius);
              fprintf (drvui->fpoutp,
                   "  texture{pigment{color %s  }}\n", edgecolor);
              fprintf(drvui->fpoutp," }\n");
              for (m=0; m<3; m++) {
                at[m] = s_vert[3 * imk + m];
                df[m] = s_vert[3 * jmk + m] - s_vert[3 * imk + m];
              }
              push_cylinder(df, at, radius, color); 
              fprintf (drvui->fpoutp,
                   " cylinder { <%f,%f,%f>,<%f,%f,%f>,%f\n",
                   s_vert[3 * jmk], s_vert[3 * jmk + 1],
                   s_vert[3 * jmk + 2], s_vert[3 * kmk],
                   s_vert[3 * kmk + 1], s_vert[3 * kmk + 2], radius);
              fprintf (drvui->fpoutp,
                   "  texture{pigment{color %s  }}\n", edgecolor);
              fprintf(drvui->fpoutp," }\n");
              for (m=0; m<3; m++) {
                at[m] = s_vert[3 * jmk + m];
                df[m] = s_vert[3 * kmk + m] - s_vert[3 * jmk + m];
              }
              push_cylinder(df, at, radius, color); 
              fprintf (drvui->fpoutp,
                   " cylinder { <%f,%f,%f>,<%f,%f,%f>,%f \n",
                   s_vert[3 * kmk], s_vert[3 * kmk + 1],
                   s_vert[3 * kmk + 2], s_vert[3 * imk],
                   s_vert[3 * imk + 1], s_vert[3 * imk + 2], radius);
              fprintf (drvui->fpoutp,
                   "  texture{pigment{color %s  }}\n", edgecolor);
              fprintf(drvui->fpoutp," }\n");
              for (m=0; m<3; m++) {
                at[m] = s_vert[3 * kmk + m];
                df[m] = s_vert[3 * imk + m] - s_vert[3 * kmk + m];
              }
              push_cylinder(df, at, radius, color); 
            }
          }
        } else {
          no[2] = vertex_list[k];         /* third point */
          jplane = 3;
          for (l = 0; l <= 2; ++l)
            v2[l] = s_vert[3 * no[2] + l] - p[l];    /* get v2 */
          sign_z = 0.0f;
          ind = 0;
          v3[0] = v1[1] * v2[2] - v1[2] * v2[1];
          v3[1] = v1[2] * v2[0] - v1[0] * v2[2];
          v3[2] = v1[0] * v2[1] - v1[1] * v2[0];
          for (l = 0; l < numb_list; ++l) {
            vol = 0.0f;
            for (m = 0; m <= 2; ++m)
              vol = vol + (s_vert[3 * vertex_list[l] + m] - p[m]) * v3[m];
            if ((float) fabs (vol) < drvui->polylimit) {
              if (l > k) {
                no[jplane++] = vertex_list[l];     /* add to plane list */
              } else {
                if ((l != i) && (l != j) && (l != k)) {
                  ind = 1;             /* we already have this plane */
                  l = numb_list;         /* skip rest of loop */
                }
              }
            } else {
/* point is not in plane - check if first out-of-plane point found */
              if (sign_z == 0.0) {
                sign_z = vol;             /* yes - set sign */
              } else {
/* if here, not first out of plane - check that vol and sign_z have same sign. 
   If not, set ind=1 to indicate not a plane */
                if (vol * sign_z < 0.0) {
                  ind = 1;
                  l = numb_list;         /* skip rest of inner loop */
                }
              }
            }
          }                	 /* end of loop on l */
          if (ind == 0) {             /* 0 means a surface plane */
            poly_list[draw_list++] = jplane;     /* number of vertices in plane */
            for (l = 0; l < jplane; ++l)
              poly_list[draw_list++] = no[l] + 1;
            fprintf (drvui->fpoutp, "/* %d-sided polygon */\n", jplane);
/* Make vertices consecutive - find center of polygon */
            xm = ym = zm = 0.0f;
            for (ii = 0; ii < jplane; ii++) {
              xm = xm + s_vert[3 * no[ii]];
              ym = ym + s_vert[3 * no[ii] + 1];
              zm = zm + s_vert[3 * no[ii] + 2];
            }
            xm = xm / (float) jplane;
            ym = ym / (float) jplane;
            zm = zm / (float) jplane;

/* sort vertices in consecutive order based on minimal angles between vectors */

            for (ii = 0; ii < jplane; ii++)
              ns[ii] = 0;
            ns[0] = no[0];
            for (ii = 0; ii < jplane - 1; ii++) {
              phi0 = 1000.0f;
              vxs = s_vert[3 * ns[ii]] - xm;
              vys = s_vert[3 * ns[ii] + 1] - ym;
              vzs = s_vert[3 * ns[ii] + 2] - zm;
              bms = (float) sqrt (vxs * vxs + vys * vys + vzs * vzs);
              for (jj = 0; jj < jplane; jj++) {
                for (ll = 0; ll < ii; ll++) {
                  if (ns[ll] == no[jj])
                    jj++;
                  if (jj >= jplane)
                    break;
                }
                if (jj < jplane) {
                  if (ns[ii] != no[jj]) {
                    vx = s_vert[3 * no[jj]] - xm;
                    vy = s_vert[3 * no[jj] + 1] - ym;
                    vz = s_vert[3 * no[jj] + 2] - zm;
                    bm = (float) sqrt (vx * vx + vy * vy + vz * vz);
                    cosphi = (vxs * vx + vys * vy + vzs * vz) / (bms * bm);
                    if (cosphi < -1.) cosphi = -1.;
                    if (cosphi > 1.) cosphi = 1.;
                    phi = acos (cosphi) * RAD;
                    if (phi < phi0) {
                      phi0 = (float) phi;
                      ns[ii + 1] = no[jj];
                    }
                  }
                }
              }
            }
            if (numb_list != 4) {
              draw_list = draw_list - jplane;
              for (l = 0; l < jplane; ++l)
                poly_list[draw_list++] = ns[l] + 1;
            }

#ifdef OPENGL
            if (gl_view) {
                glPushMatrix ();
                glLoadName (i0*100000+i0);
                glPushName (i1);
                (void)sscanf (poly_col_v, "%f %f %f", &glr, &glg, &glb);
                glDisable (GL_LIGHTING);
                glColor3f (glr, glg, glb);
                glBegin (GL_TRIANGLES);
                for (l = 0; l < jplane - 2; ++l) {
                  d1 = (float) fabs (s_vert[3 * ns[0]] - s_vert[3 * ns[l + 1]]) +
                   (float) fabs (s_vert[3 * ns[0] + 1] -  s_vert[3 * ns[l + 1] + 1]) +
                   (float) fabs (s_vert[3 * ns[0] + 2] -  s_vert[3 * ns[l + 1] + 2]);
                  d2 = (float) fabs (s_vert[3 * ns[0]] - s_vert[3 * ns[l + 2]]) +
                   (float) fabs (s_vert[3 * ns[0] + 1] -  s_vert[3 * ns[l + 2] + 1]) +
                   (float) fabs (s_vert[3 * ns[0] + 2] -  s_vert[3 * ns[l + 2] + 2]);
                  d3 = (float) fabs (s_vert[3 * ns[l + 2]] - s_vert[3 * ns[l + 1]]) +
                   (float) fabs (s_vert[3 * ns[l + 2] + 1] -  s_vert[3 * ns[l + 1] + 1]) +
                   (float) fabs (s_vert[3 * ns[l + 2] + 2] -  s_vert[3 * ns[l + 1] + 2]);
                  if ((d1 > 0.01) && (d2 > 0.01) && (d3 > 0.01)) {
                    fprintf (drvui->fpoutp, " triangle {< %8.5f, %8.5f, %8.5f>,\n",s_vert[3 * ns[0]],
                      s_vert[3 * ns[0] + 1], s_vert[3 * ns[0] + 2]);
                    fprintf (drvui->fpoutp, "<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>\n",
                      s_vert[3 * ns[l + 1]],s_vert[3 * ns[l + 1] + 1],
                      s_vert[3 * ns[l + 1] + 2],s_vert[3 * ns[l + 2]],
                      s_vert[3 * ns[l + 2] + 1],s_vert[3 * ns[l + 2] + 2]);
                    fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n",poly_col_p);
                    fprintf(drvui->fpoutp," }\n");

/* 
      need to add something like
      glNormal3f(xm-xc,ym-yc,zm-zc ); 
      for lighting -  xc,yc,zc being the center of the polyhedron
      (which is not normally available to build_poly_list)
*/
                    glVertex3f (s_vert[3 * ns[0]],s_vert[3 * ns[0] + 1], s_vert[3 * ns[0] + 2]);
                    glVertex3f (s_vert[3 * ns[l + 1]],s_vert[3 * ns[l + 1] + 1],s_vert[3 * ns[l + 1] + 2]);
                    glVertex3f (s_vert[3 * ns[l + 2]],s_vert[3 * ns[l + 2] + 1],s_vert[3 * ns[l + 2] + 2]);
                  }

                } 
                glEnd ();
                glPopName();
                glEnable (GL_LIGHTING);
                glPopMatrix ();
            }
#endif
            if (edges) {
              float at[3], df[3];
              int m;
              char color[128];
              if (drvui->p_rad_edge[polyno] >= 0.001f)
                strncpy (edgecolor, drvui->p_col_edge[polyno],39);
              else
                strncpy (edgecolor, drvui->col_edge,39);
              trim_string(edgecolor,40);
              strcpy(color, edgecolor);
              Transform_POV_Color(edgecolor);
              Transform_VRML_Color(color);
              for (l = 0; l < jplane - 1; ++l) {
                d1 = (float) fabs (s_vert[3 * ns[l]] - s_vert[3 * ns[l + 1]]) +
                     (float) fabs (s_vert[3 * ns[l] + 1] - s_vert[3 * ns[l + 1] + 1]) +
                     (float) fabs (s_vert[3 * ns[l] + 2] - s_vert[3 * ns[l + 1] + 2]);

                if (d1 > 0.01) {
                  fprintf (drvui->fpoutp," cylinder { <%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f> , %8.5f \n",
                     s_vert[3 * ns[l]],s_vert[3 * ns[l] + 1],s_vert[3 * ns[l] + 2],
                     s_vert[3 * ns[l + 1]],s_vert[3 * ns[l + 1] + 1],s_vert[3 * ns[l + 1] + 2], radius);
                  fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n", edgecolor);
                  fprintf(drvui->fpoutp," }\n");
                  for (m=0; m<3; m++) {
                    at[m] = s_vert[3 * ns[l] + m];
                    df[m] = s_vert[3 * ns[l + 1] + m] - at[m];
                  }
                  push_cylinder(df, at, radius, color);
                }
              }
              d1 = (float) fabs (s_vert[3 * ns[0]] - s_vert[3 * ns[jplane - 1]]) +
                   (float) fabs (s_vert[3 * ns[0] + 1] -  s_vert[3 * ns[jplane - 1] + 1]) +
                   (float) fabs (s_vert[3 * ns[0] + 2] -  s_vert[3 * ns[jplane - 1] + 2]);
              if (d1 > 0.01) {
                fprintf (drvui->fpoutp, " cylinder { <%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f> , %8.5f \n",
                   s_vert[3 * ns[0]],s_vert[3 * ns[0] + 1],s_vert[3 * ns[0] + 2],
                   s_vert[3 * ns[jplane - 1]],s_vert[3 * ns[jplane - 1] + 1],
                   s_vert[3 * ns[jplane - 1] + 2], radius);
                fprintf (drvui->fpoutp, "  texture{pigment{color %s  }}\n", edgecolor);
                fprintf(drvui->fpoutp," }\n");
                for (m=0; m<3; m++) {
                  at[m] = s_vert[3 * ns[0] + m];
                  df[m] = s_vert[3 * ns[jplane - 1] + m] - at[m];
                }
                push_cylinder(df, at, radius, color);
              }
            }
          }
        }                     /* end of numb_list != 4 */
      }                    	 /* end of for (k.... */
    }
  }
  if (c1) {
      fprintf(drvui->flout," %7.3f %8.5f %8.5f %8.3f\n",Vol, sigd, qe, siga);
  } else {
      fprintf(drvui->flout," %7.3f %8.5f\n",Vol, sigd);
  }
  free (no);
  free (ns);
  return;
}                    	 /* end of build_poly_list */

/* ************************************************************** */
/* ************************************************************** */

void crystal_compute (int Quick, int keep_matrix)
{
  int i;
  char outfil[100];
  float temp;
  float rm[16];
  char col_bg_v[40];
  char col_bg_p[40];
  float o_xrot=0.,o_yrot=0.,o_zrot=0.;

  if (drvui->Str_File == NULL)
    return;					 /* Fail immediately if filename NULL */
/* initialize */
  origin[0] = drvui->X_Origin;
  origin[1] = drvui->Y_Origin;
  origin[2] = drvui->Z_Origin;
  boxlim[0] = drvui->X_Boxlim;
  boxlim[1] = drvui->Y_Boxlim;
  boxlim[2] = drvui->Z_Boxlim;
  docell = Unit_Cell;
  do_ellipsoids = Color_Warning = 0;
  rad_cell = 0.02f;
  Sphere_Mult = SpMult;
  drvui->Bond_Mult = drvui->BndMult;
  drvui->Str_File_Inp = drvui->Str_File;
  natom = 0;
  drvui->nmag = 0;
  drvui->nsphere = drvui->npoly = drvui->nbond = drvui->nplane = drvui->nbplane = 1;
  drvui->n_ellips = drvui->ncone = drvui->nlabel = 1;
  drvui->poly_size[0] = 2.5f;
  drvui->poly_min[0] = 0.0f;
  drvui->poly_l[0][0] = drvui->poly_l[0][1] = drvui->poly_l[0][2] = drvui->poly_l[0][3] = ' ';
  drvui->plane_size[0] = 2.5f;
  drvui->plane_l[0][0] = drvui->plane_l[0][1] = drvui->plane_l[0][2] = drvui->plane_l[0][3] = ' ';
  drvui->sphere_size[0] = 0.0f;
  drvui->sphere_l[0][0] = drvui->sphere_l[0][1] = drvui->sphere_l[0][2] = drvui->sphere_l[0][3] = ' ';
  POV_Max[0] = POV_Max[1] = POV_Max[2] = -99999.0f;
  POV_Min[0] = POV_Min[1] = POV_Min[2] = 99999.0f;
  strcpy (drvui->col_bg, "White");
  drvui->frame_no = 0;
  drvui->max_frame = 99999;
  drvui->ambient = 0.0f;                   /* POV lighting options */
  drvui->diffuse = 0.0f;
  drvui->specular = 0.0f;
  drvui->roughness = 0.0f;
#if defined(__APPLE__) && defined(__MACH__)    /* reset working directory in OS X */
  i = chdir(drvui->savedpath);
  if (i != 0) printf ("Warning: change of working directory to %s failed.\n",drvui->savedpath);
#endif
  strcpy (outfil, drvui->Str_File);
  for (i = (int) strlen (outfil); i>0; --i) {
    if (outfil[i] == '.' || outfil[i] == ' ')
      break;
  }
  outfil[i] = '\0';
  strcpy(drvui->File_Root,outfil);
  if (!Quick) {
    strcat (outfil, ".out");
    if ((drvui->flout = fopen (outfil, "w")) == NULL) {
      fprintf (stderr, "Cannot open listing file\n");
      delay (5);
      return;
    }
  }
  if (!Quick) {
    outfil[i] = '\0';
    strcat (outfil, ".wrl");
    if ((drvui->fpoutv = fopen (outfil, "w")) == NULL) {
      fprintf (stderr, "Cannot open VRML output file\n");
      delay (5);
      return;
    }
  }
  if (!Quick) {
    outfil[i] = '\0';
    strcpy(drvui->Pov_File,outfil);
    strcat (drvui->Pov_File, ".pov");
    if ((drvui->fpoutp = fopen (drvui->Pov_File, "w")) == NULL) {
      fprintf (stderr, "Cannot open POV output file\n");
      delay (5);
      return;
    }
  }

  if (!Quick) {
    fprintf (drvui->flout, "\n\n                DRAWxtl\n");
    fprintf (drvui->flout, "  Crystal Structure Drawing Program V%s\n\n", VERSION);
    fprintf (drvui->flout,
	     "  Copyright 1996, 1997, 2000-2007 by Larry Finger, Martin Kroeker, and Brian Toby\n");
    fprintf (drvui->flout, "  (all rights reserved).\n\n");
    fprintf (drvui->flout, "     Combined POVxtl and VRMLxtl\n");
    fprintf (drvui->flout,
	     "POVxtl  - Scenes for Persistence of Vision Ray Tracer (POV-Ray) V2.0\n");
    fprintf (drvui->flout,
	     "VRMLxtl - Virtual Reality Modeling Language V1.0/V2.0\n");
    getcwd(outfil,100);
    fprintf (drvui->flout, 
             "DRAWxtl working directory is %s\n",outfil);
    printf ("\n\n                DRAWxtl\n");
    printf ("  Crystal Structure Drawing Program V%s\n\n", VERSION);
    printf ("  Copyright 1996, 1997, 2000-2007 by Larry Finger, Martin Kroeker, and Brian Toby\n");
    printf ("  (all rights reserved).\n\n");
    printf ("     Combined POVxtl and VRMLxtl\n");
    printf ("POVxtl  - Scenes for Persistence of Vision Ray Tracer (POV-Ray) V2.0\n");
    printf ("VRMLxtl - Virtual Reality Modeling Language V1.0/V2.0\n");
    getcwd(outfil,100);
    printf ("DRAWxtl working directory is %s\n",outfil);
  }
#ifdef OPENGL
  if (gl_view) {
      glClear ((GLbitfield) (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT));
      if (!glIsList(crystalDL)) crystalDL = glGenLists (1);
      glNewList (crystalDL, GL_COMPILE);
  } 
#endif
  while (drvui->frame_no != drvui->max_frame) {             /* loop through all frames */
    drvui->frame_no++;
    if (!Quick) {
      fprintf (drvui->flout, "\n doing frame %d\n",drvui->frame_no);
      fprintf (stderr, " doing frame %d\n",drvui->frame_no);
    }
    if (drvui->frame_no == 1) {
        clipflag = 0;
        packflag = 0;
        domolcomp = 0;
        origin[0] = drvui->X_Origin;
        origin[1] = drvui->Y_Origin;
        origin[2] = drvui->Z_Origin;
        docell = Unit_Cell;
        do_ellipsoids = Color_Warning = 0;
        drvui->noshadow = 0;
        rad_cell = 0.02f;
        Sphere_Mult = SpMult;
        drvui->Bond_Mult = drvui->BndMult;
    }
    boxlim[0] = drvui->X_Boxlim;
    boxlim[1] = drvui->Y_Boxlim;
    boxlim[2] = drvui->Z_Boxlim;
    natom = 0;
    if ((drvui->fpin = fopen (drvui->Str_File, "r")) == NULL) {
      printf ("Cannot Open Input File");
      return;
    }
    if (keep_matrix == 1) {
     o_xrot=xrot;
     o_yrot=yrot;
     o_zrot=zrot;
    } 
    get_input (Quick);
    if (keep_matrix == 1) {
      xrot=o_xrot;
      yrot=o_yrot;
      zrot=o_zrot;
    } 
    fclose(drvui->fpin);
    strcpy(col_bg_v,drvui->col_bg);
    Transform_VRML_Color(col_bg_v);
    strcpy(col_bg_p,drvui->col_bg);
    Transform_POV_Color(col_bg_p);
    sscanf(col_bg_v,"%f %f %f",&drvui->glback[0],&drvui->glback[1],&drvui->glback[2]);

    if (boxlim[0] == 100.0) {
      boxlim[0] = (drvui->cryst_lim[drvui->frame_no][3]
                - drvui->cryst_lim[drvui->frame_no][0]) * drvui->lat_con[0];
      boxlim[1] = (drvui->cryst_lim[drvui->frame_no][4]
                - drvui->cryst_lim[drvui->frame_no][1]) * drvui->lat_con[1];
      boxlim[2] = (drvui->cryst_lim[drvui->frame_no][5]
                - drvui->cryst_lim[drvui->frame_no][2]) * drvui->lat_con[2];
    }

/* Calculate Grand rotation matrix */
    if (drvui->frame_no == 1  && keep_matrix == 0 ) {
        int m;
        tb_init();
        XYZ_Rot_to_Q(xrot, yrot, zrot);
        quaternion_to_rotmatrix(&Rotq, rm);  /* get rotation matrix */
        for (m = 0; m < 3; m++) {            /*  and copy to G_Rot */
          G_Rot[0][m] = rm[m];
          G_Rot[1][m] = rm[m + 4];
          G_Rot[2][m] = rm[m+8];
        }

/* Initialize POV output limits */

        POV_Max[0] = POV_Max[1] = POV_Max[2] = -99999.0f;
        POV_Min[0] = POV_Min[1] = POV_Min[2] =  99999.0f;
    }
/* generate all possible atoms to get the size of the drawing */

    NvertM = 1;                      /* make master list empty */
    if (!s_vert) {
      if ((s_vert = zalloc ((unsigned) (3 * (long) 
          (MAX_VERTS * sizeof (float))))) ==	NULL) {
        printf ("Unable to get initial s_vert allocation\n");
        delay (5);
        return;
      }
    }  
    if (drvui->vert_occ) free (drvui->vert_occ);
    if (!(drvui->vert_occ = zalloc ((unsigned) (MAX_VERTS * sizeof (float))))) {
      printf ("Unable to get site occupancy allocation\n");
      delay(5);
      return;
    }
    memset(drvui->vert_occ, 0, (MAX_VERTS * sizeof (float)));
    get_atom_id ();                  /* rework atom id numbers to pointers */
    if (slabmode) generate_slab();      /* initialize cutout box if needed */
    if (drvui->modulated > 0) {
        float save_limits[6];
        int j, atom;
        fprintf(drvui->flout,"\nTransformed (modulated) contents of 2x2x2 unit-cells corresponding to\n"
                           " the 'simulated' P1 positions of JANA 2000\n\n"
                           "  Atom       x'       y'       z'   site occ.   Sym. Op\n\n");
        for (i=0; i<3; i++) {
            save_limits[i] = drvui->cryst_lim[drvui->frame_no][i];
            drvui->cryst_lim[drvui->frame_no][i] = -1.0f;  /* set limits from -1 to +1 in all directions */
            save_limits[i+3] = drvui->cryst_lim[drvui->frame_no][i+3];
            drvui->cryst_lim[drvui->frame_no][3+i] = 1.0f;
        }
        build_box_contents ();              /*  get all atoms in the 2x2x2 region */
        for (i=0; i<6; i++) {
            drvui->cryst_lim[drvui->frame_no][i] = save_limits[i];  /* restore original limits */
        }
        for (atom=0; atom<natom; atom++) {
            for (j=1; j<NvertM; j++) {
                int sym, k, m;

                if (atom != drvui->atom_no[j]/100) continue;
                sym = drvui->atom_no[j] % 100;
                k = 0;
                for (m=0; m<3; m++) {
                    if (xypos[3*j+m] < -1.0f) k = 1;
                    if (xypos[3*j+m] >  1.0f) k = 1;
                }
                if (!k)
                    fprintf(drvui->flout," %c%c%c%c%3d %8.5f %8.5f %8.5f %8.5f %7d\n",
                            drvui->atom_l[atom][0], drvui->atom_l[atom][1], drvui->atom_l[atom][2],
                            drvui->atom_l[atom][3], drvui->sv_atom_n[atom], xypos[3*j], xypos[3*j+1],
                            xypos[3*j+2], drvui->vert_occ[j], sym+1);
            }
        }
        NvertM = 1;                     /* make master list empty and reset limits */
        POV_Max[0] = POV_Max[1] = POV_Max[2] = -99999.0f;
        POV_Min[0] = POV_Min[1] = POV_Min[2] = 99999.0f;
    }
    build_box_contents ();           /* complete the master atom list */
/*    free (s_vert);*/
    printf ("\nMaster atom list contains %5d atoms.\n", NvertM - 1);
    fprintf (drvui->flout, "\nMaster atom list contains %5d atoms.\n", NvertM - 1);
/* allocate space for the various arrays */
    if (o_vert)
      free(o_vert);
    if ((o_vert = 
	 zalloc ((unsigned) (6 * (NvertM + 10) * sizeof (float)))) == NULL) {
      printf ("Unable to get o_vert allocation\n");
      delay (5);
      return;
    }
    if (o_vert_nm)
        free(o_vert_nm);
    if (!(o_vert_nm = (float *)zalloc ((unsigned) (6 * (NvertM+20) *    sizeof (float))))) {
      printf ("Unable to get o_vert_nm allocation\n");
      delay(5);
      return;
    }
    if (vert_sym_no)
      free(vert_sym_no);
    if ((vert_sym_no =
	 zalloc ((unsigned) ((2 * NvertM + 10) * sizeof (int)))) ==	NULL) {
      printf ("Unable to get vert_sym_no allocation\n");
      delay (5);
      return;
    }
    if (drvui->orig_atom_no)
      free(drvui->orig_atom_no);
    if ((drvui->orig_atom_no =
	 zalloc ((unsigned) ((2 * NvertM + 10) * sizeof (int)))) ==	NULL) {
      printf ("Unable to get orig_atom_no allocation\n");
      delay (5);
      return;
    }
    if ((poly_list =
	 zalloc ((unsigned) ((4 * (NvertM + 10)) * sizeof (int)))) == NULL) {
      printf ("Unable to get poly_list allocation\n");
      delay (5);
      return;
    }
    if ((vertex_list =
	 zalloc ((unsigned) ((4 * (NvertM + 10)) * sizeof (int)))) == NULL) {
      printf ("Unable to get vertex_list allocation\n");
      delay (5);
      return;
    }

    if (!drvui->subsys_vol[0]) 
      drvui->subsys_vol[0] = drvui->subsys_ref_volume;

    analyze_bonds ();				 /* print the bond distance table */
    printf ("\nThe 'pov' and 'wrl' files contain:\n");
    if (docell != 0 || natom == 0) {      /* if unit-cell corners drawn, add them for scaling */
        float vert[3];
        int saved_Nvert;
        int nocell;

        for (nocell=0; nocell<drvui->no_subsys; nocell++) {
            saved_Nvert = NvertM;
            for (i = 0; i <= 2; ++i)
                vert[i] = 0.0f;
            sub_add_vert (vert, nocell);             /* place 0,0,0 in list */
            vert[2] = 1.0f;
            sub_add_vert (vert, nocell);             /* place 0,0,1 in list */
            vert[1] = 1.0f;
            sub_add_vert (vert, nocell);             /* place 0,1,1 in list */
            vert[2] = 0.0f;
            sub_add_vert (vert, nocell);             /* place 0,1,0 in list */
            vert[0] = 1.0f;
            sub_add_vert (vert, nocell);             /* place 1,1,0 in list */
            vert[1] = 0.0f;
            sub_add_vert (vert, nocell);             /* place 1,0,0 in list */
            vert[2] = 1.0f;
            sub_add_vert (vert, nocell);             /* place 1,0,1 in list */
            vert[1] = 1.0f;
            sub_add_vert (vert, nocell);             /* place 1,1,1 in list */
            NvertM = saved_Nvert;              /* reset list count */
        }
    }
    if (drvui->frame_no == 1) {			      /* only do for first frame */
      temp = max(POV_Max[0] - POV_Min[0],POV_Max[1] - POV_Min[1]);
      temp = max(temp, POV_Max[2] - POV_Min[2]);
      if (gl_size == 99999.0f) gl_size = temp;   /* only do ONCE */
      Scale = 15.0f * temp / Magnification;

      if (Vrml2)
        fprintf (drvui->fpoutv, "#VRML V2.0 utf8\n\n");
      else
        fprintf (drvui->fpoutv, "#VRML V1.0 ascii\n\n");
      fprintf (drvui->fpoutv, "# Scene file created by DRAWxtl V%s from %s\n",
	       VERSION, drvui->Str_File_Inp);
      fprintf (drvui->fpoutv,
	       "# with command line options: -b %6.2f %6.2f %6.2f -o %5.2f %5.2f %5.2f\n",
	       boxlim[0], boxlim[1], boxlim[2], origin[0], origin[1],
	       origin[2]);
      fprintf (drvui->fpoutv,
	       "#  -p %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f -v %5.2f %5.2f %5.2f\n",
	       drvui->cryst_lim[drvui->frame_no][0], drvui->cryst_lim[drvui->frame_no][3],
           drvui->cryst_lim[drvui->frame_no][1], drvui->cryst_lim[drvui->frame_no][4],
           drvui->cryst_lim[drvui->frame_no][2], drvui->cryst_lim[drvui->frame_no][5],
           xrot, yrot, zrot);

/* Output VRML header */

      if (!Vrml2) {
        fprintf (drvui->fpoutv,"DEF BackgroundColor Info { string \"%s\" }\n", col_bg_v);
        fprintf (drvui->fpoutv, "Separator {\n");
        fprintf (drvui->fpoutv, " Material {\n");
        fprintf (drvui->fpoutv, "  ambientColor    0.1 0.1 0.1\n");
        fprintf (drvui->fpoutv, "  diffuseColor    1 1 1\n");
        fprintf (drvui->fpoutv, "  specularColor   0.9 0.9 0.9\n");
        fprintf (drvui->fpoutv, "  shininess       0.8\n");
        fprintf (drvui->fpoutv, " }\n");
      }
      Text_Size = 0.0005f *Scale;

      if (!M_cameras && Vrml2) {		 /* Generate Orthographic Projection for VRML97 */
        fprintf (drvui->fpoutv, "NavigationInfo {type \"EXAMINE\"}\n");
        fprintf (drvui->fpoutv, "Background { skyColor %s }\n", col_bg_v);
        fprintf (drvui->fpoutv,
		 "Viewpoint { fieldOfView 0.1 position 0. 0. %8.3f }\n", Scale);
        fprintf (drvui->fpoutv, "DirectionalLight {direction 0 0 -1}\n");
        fprintf (drvui->fpoutv, "DirectionalLight {direction 0 0 1}\n");
      } else {
        if (!Vrml2) {				 /* simulate orthographic projection for VRML1.0 */
          float fscale, fangle;

          if (!M_cameras) {
            fscale = Scale * (10.0f / 7.0f);
            fangle = 0.07853f;
          } else {
            fscale = Scale / 7.0f;
            fangle = 0.7853f;
          }
          fprintf (drvui->fpoutv, "PerspectiveCamera {\n");
          fprintf (drvui->fpoutv, "  position        0 0 %8.3f\n", fscale);
          fprintf (drvui->fpoutv, "  orientation     0 0 1 0\n");
          fprintf (drvui->fpoutv, "  focalDistance %8.3f\n", fscale);
          fprintf (drvui->fpoutv, "  heightAngle   %8.5f\n", fangle);
          fprintf (drvui->fpoutv, "}\n");
          fprintf (drvui->fpoutv,
		   "DirectionalLight{ direction 0 0 -1 intensity 1.0}\n");
          fprintf (drvui->fpoutv,
		   "DirectionalLight{ direction 0 0 1 intensity 1.0}\n");
        } else {
          fprintf (drvui->fpoutv, "NavigationInfo {type \"EXAMINE\"}\n");
          fprintf (drvui->fpoutv, "Background { skyColor %s }\n", col_bg_v);
          fprintf (drvui->fpoutv, "Viewpoint { position 0. 0. %8.3f }\n", Scale / 7.);
          fprintf (drvui->fpoutv, "DirectionalLight {direction 0 0 -1}\n");
          fprintf (drvui->fpoutv, "DirectionalLight {direction 0 0 1}\n");
        } 
      }

      if (Vrml2) {
        fprintf (drvui->fpoutv, "Transform {");
        fprintf (drvui->fpoutv, " rotation 0 0 1 %f\n", zrot / RAD);
        fprintf (drvui->fpoutv, " children Transform {");
        fprintf (drvui->fpoutv, " rotation 0 1 0 %f\n", yrot / RAD);
        fprintf (drvui->fpoutv, " children Transform {");
        fprintf( drvui->fpoutv, " rotation 1 0 0 %f\n", xrot / RAD);
        fprintf (drvui->fpoutv, "  children [\n");
      } else {
        fprintf (drvui->fpoutv, "Rotation {");
        fprintf (drvui->fpoutv, " rotation 0 0 1 %f}\n", zrot / RAD);
        fprintf (drvui->fpoutv, " Rotation {");
        fprintf (drvui->fpoutv, " rotation 0 1 0 %f}\n", yrot / RAD);
        fprintf (drvui->fpoutv, " Rotation {");
        fprintf (drvui->fpoutv, " rotation 1 0 0 %f}\n", xrot / RAD);
      } 

/* Output POV header */

        fprintf (drvui->fpoutp, "/*\n\n    POVxtl-Persistence of Vision Ray Tracer\n");
        fprintf (drvui->fpoutp, "    Crystal Structure Drawing Program %s\n\n", VERSION);
        fprintf (drvui->fpoutp,
	       "       Copyright 1996, 1997, 2000-2007 by Larry Finger, Martin Kroeker, and Brian Toby\n");
        fprintf (drvui->fpoutp, "          (all rights reserved).\n\n");
        fprintf (drvui->fpoutp, " Scene file created from %s \n", drvui->Str_File_Inp);
        fprintf (drvui->fpoutp,
	       " with command line options: -b %6.2f %6.2f %6.2f -o %5.2f %5.2f %5.2f \n",
	       boxlim[0], boxlim[1], boxlim[2], origin[0], origin[1],
	       origin[2]);
        fprintf (drvui->fpoutp,
	       "  -p %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f -v %5.2f %5.2f %5.2f -m %5.2f\n*/\n\n",
	       drvui->cryst_lim[drvui->frame_no][0], drvui->cryst_lim[drvui->frame_no][3],
           drvui->cryst_lim[drvui->frame_no][1], drvui->cryst_lim[drvui->frame_no][4],
           drvui->cryst_lim[drvui->frame_no][2], drvui->cryst_lim[drvui->frame_no][5],
           xrot, yrot, zrot, Magnification);

        if (drvui->Stereo)
          fprintf(drvui->fpoutp,"#version unofficial stereopov 0.2\n");
        fprintf (drvui->fpoutp, "#include \"colors.inc\"\n");
        fprintf (drvui->fpoutp, "#include \"chars.inc\"\n");
        if (drvui->roughness > 0.) {
          fprintf (drvui->fpoutp, "\ndefault {\n");
          fprintf (drvui->fpoutp, "   finish {\n");
          fprintf (drvui->fpoutp, "     ambient %5.2f\n",drvui->ambient);
          fprintf (drvui->fpoutp, "     diffuse %5.2f\n",drvui->diffuse);
          fprintf (drvui->fpoutp, "     specular %5.2f\n",drvui->specular);
          fprintf (drvui->fpoutp, "     roughness %5.2f\n",drvui->roughness);
          fprintf (drvui->fpoutp, "   }\n");
          fprintf (drvui->fpoutp, " }\n");
        }	
        if (!M_cameras) {
          fprintf (drvui->fpoutp, "camera{ orthographic\n");
        } else {
          fprintf (drvui->fpoutp, "camera{\n");
      }
      Scale *= 0.9f;
      fprintf (drvui->fpoutp, " location< 0 ,0, %f>\n", 1.6 * Scale);
      if (!M_cameras) {
        fprintf (drvui->fpoutp, " up <0,%f,0>\n", 0.1 * Scale);
        fprintf (drvui->fpoutp, " right <%f,0,0>\n", -0.1 * Scale);
      } else {
        fprintf (drvui->fpoutp, " up <0, 0.1, 0>\n");
        fprintf (drvui->fpoutp, " right <-0.1, 0, 0>\n");
      }
      fprintf (drvui->fpoutp, " look_at <0,0,0>\n");
      if (drvui->Stereo) {
          fprintf(drvui->fpoutp," stereo_base %f\n",drvui->stereo_base);
          if (drvui->cross_eyed)
              fprintf(drvui->fpoutp," cross_eyed\n");
      }
      fprintf (drvui->fpoutp, "}\n");
      fprintf (drvui->fpoutp, "background{color %s}\n", col_bg_p);
      fprintf (drvui->fpoutp, "light_source {< 0, 0, %f>\n", Scale);
      fprintf (drvui->fpoutp, " color red 4.0 green 4.0 blue 4.0\n}\n");
      fprintf (drvui->fpoutp, "object{\n union{\n");
      fprintf (drvui->fpoutp, "\n");
      fprintf (drvui->fpoutp, "#declare xrot = %5.1f;\n", xrot);
      fprintf (drvui->fpoutp, "#declare yrot = %5.1f;\n", yrot);
      fprintf (drvui->fpoutp, "#declare zrot = %5.1f;\n", zrot);
      fprintf (drvui->fpoutp, "\n");
    } else {             /* stuff for following frames */
      fprintf (drvui->fpoutp, "\n/************ Frame %d **********/\n", drvui->frame_no);
      fprintf (drvui->fpoutv, "\n############# Frame %d ##########\n", drvui->frame_no);
    }    /* only do for first frame */
    do_drawing ();				 /* generate all elements of the drawing */
    free (vertex_list);
    free (poly_list);
  }						 /* while (frame_no != max_frame) - loop through all frames */

#ifdef OPENGL
  if (gl_view)
      glEndList ();
#endif
  fprintf (drvui->fpoutp, " }\n");
  fprintf (drvui->fpoutp, "rotate <xrot, 0, 0>\n");
  fprintf (drvui->fpoutp, "rotate <0, yrot, 0>\n");
  fprintf (drvui->fpoutp, "rotate <0, 0, zrot>\n");
  fprintf (drvui->fpoutp, "}\n");
  (void) fflush (drvui->fpoutp);

  if (Vrml2)
    fprintf (drvui->fpoutv, "]\n");
  fprintf (drvui->fpoutv, " }\n");
  if (Vrml2) {
    fprintf (drvui->fpoutv, " }\n");
    fprintf (drvui->fpoutv, " }\n");
  }

  (void) fflush (drvui->flout);
  (void) fflush (drvui->fpoutv);
}						 /* end of crystal_compute */

/* ************************************************************** */
/* ************************************************************** */

void draw_cell (int docell)

/* procedure to add unit-cell outline to geometry file */
/* docell - non-zero if unit-cell to be drawn */
{
    int i,j,k;
    float vert[3];
    float radius;
    float df[3];
    float at[3];
    char col_cell_v[40];
    char col_cell_p[40];
    int nocell;

    radius = rad_cell * 0.01f *Scale;

    if (drvui->no_subsys==1)    
        drvui->subsys_vol[0] = drvui->subsys_ref_volume;

    for (nocell=0; nocell<drvui->no_subsys; nocell++) {
        nvert = 0;                     /* initialize vertex list */
        for (i = 0; i <= 2; ++i)
            vert[i] = 0.0f;
        sub_add_vert_nc (vert, nocell);             /* place 0,0,0 in list */
        vert[2] = 1.0f;
        sub_add_vert_nc (vert, nocell);             /* place 0,0,1 in list */
        vert[1] = 1.0f;
        sub_add_vert_nc (vert, nocell);             /* place 0,1,1 in list */
        vert[2] = 0.0f;
        sub_add_vert_nc (vert, nocell);             /* place 0,1,0 in list */
        vert[1] = 0.0f;
        sub_add_vert_nc (vert, nocell);             /* place 0,0,0 in list */
        vert[0] = 1.0f;
        sub_add_vert_nc (vert, nocell);             /* place 1,0,0 in list */
        vert[1] = 1.0f;
        sub_add_vert_nc (vert, nocell);             /* place 1,1,0 in list */
        vert[2] = 1.0f;
        sub_add_vert_nc (vert, nocell);             /* place 1,1,1 in list */
        vert[1] = 0.0f;
        sub_add_vert_nc (vert, nocell);             /* place 1,0,1 in list */
        vert[2] = 0.0f;
        sub_add_vert_nc (vert, nocell);             /* place 1,0,0 in list */
        vert[1] = 1.0f;
        sub_add_vert_nc (vert, nocell);             /* place 1,1,0 in list */
        vert[0] = 0.0f;
        sub_add_vert_nc (vert, nocell);             /* place 0,1,0 in list */
        vert[2] = 1.0f;
        sub_add_vert_nc (vert, nocell);             /* place 0,1,1 in list */
        vert[0] = 1.0f;
        sub_add_vert_nc (vert, nocell);             /* place 1,1,1 in list */
        vert[1] = 0.0f;
        sub_add_vert_nc (vert, nocell);             /* place 1,0,1 in list */
        vert[0] = 0.0f;
        sub_add_vert_nc (vert, nocell);             /* place 0,0,1 in list */

        if (docell) {
            strcpy(col_cell_v,drvui->col_cell);
            Transform_VRML_Color(col_cell_v);
            strcpy(col_cell_p,drvui->col_cell);
            Transform_POV_Color(col_cell_p);
            if (Vrml2) {
                fprintf (drvui->fpoutv, " Shape {\n");
                fprintf (drvui->fpoutv,"  geometry IndexedLineSet{ coord Coordinate{ point[\n");
            } else {
                fprintf (drvui->fpoutv, " Separator{\n");
                fprintf (drvui->fpoutv, "  Material{\n");
                fprintf (drvui->fpoutv, "   diffuseColor %s\n", col_cell_v);
                fprintf (drvui->fpoutv, "  }\n");
                fprintf (drvui->fpoutv, "  Coordinate3{ point[\n");
            }
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[0], s_vert[1],s_vert[2]);
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[3], s_vert[4],s_vert[5]);
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[6], s_vert[7],s_vert[8]);
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[9], s_vert[10],s_vert[11]);
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[15], s_vert[16],s_vert[17]);
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[18], s_vert[19],s_vert[20]);
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[21], s_vert[22],s_vert[23]);
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f]\n  }\n", s_vert[24],s_vert[25], s_vert[26]);
            if (Vrml2) {
                fprintf (drvui->fpoutv, "  coordIndex[");
            } else {
                fprintf (drvui->fpoutv, "  IndexedLineSet{coordIndex[");
            }
            fprintf (drvui->fpoutv, "0,1,2,3,0,4,5,6,7,1,2,6,5,3,0,4,7,-1]\n");
            if (Vrml2) {
                fprintf (drvui->fpoutv, "  color Color { color [%s]}\n  colorIndex[0]\n"
                   "  colorPerVertex FALSE\n",col_cell_v);
            }
            fprintf (drvui->fpoutv, " }\n");
            fprintf (drvui->fpoutv, "}\n");

            for (i=0;i<12;i++) {                 /* generate unit-cell edges for openGL */
                j = i;
                if (i == 9) j = 10;
                if (i == 10) j = 12;
                k = j+1;
                if (i == 11) {
                    j = 8;
                    k = 1;
                }
                at[0] = s_vert[3*j];              /* beginning vertex */
                at[1] = s_vert[3*j+1];
                at[2] = s_vert[3*j+2];
                df[0] = s_vert[3*k] - at[0];      /* difference from start to end */
                df[1] = s_vert[3*k+1] - at[1];
                df[2] = s_vert[3*k+2] - at[2];
                push_cylinder(df,at, radius, col_cell_v);  /* draw the cylinder */
            }
            fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f \n",
                    s_vert[0], s_vert[1], s_vert[2], s_vert[3], s_vert[4],
                    s_vert[5], radius);
            fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
            fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
            fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
                    s_vert[3], s_vert[4], s_vert[5], s_vert[6], s_vert[7],
                    s_vert[8], radius);
            fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
            fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
            fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
                    s_vert[6], s_vert[7], s_vert[8], s_vert[9], s_vert[10],
                    s_vert[11], radius);
            fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
            fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
            fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>, %f\n",
                    s_vert[9], s_vert[10], s_vert[11], s_vert[12], s_vert[13],
                    s_vert[14], radius);
            fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
            fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
            fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
                    s_vert[12], s_vert[13], s_vert[14], s_vert[15], s_vert[16],
                    s_vert[17], radius);
            fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
            fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
            fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
                    s_vert[15], s_vert[16], s_vert[17], s_vert[18], s_vert[19],
                    s_vert[20], radius);
            fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
            fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
            fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
                    s_vert[18], s_vert[19], s_vert[20], s_vert[21], s_vert[22],
                    s_vert[23], radius);
            fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
            fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
            fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
                    s_vert[21], s_vert[22], s_vert[23], s_vert[24], s_vert[25],
                    s_vert[26], radius);
            fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
            fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
            fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
                    s_vert[24], s_vert[25], s_vert[26], s_vert[27], s_vert[28],
                    s_vert[29], radius);
            fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
            fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
            fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
                    s_vert[30], s_vert[31], s_vert[32], s_vert[33], s_vert[34],
                    s_vert[35], radius);
            fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
            fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
            fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
                    s_vert[36], s_vert[37], s_vert[38], s_vert[39], s_vert[40],
                    s_vert[41], radius);
            fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
            fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
            fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
                    s_vert[24], s_vert[25], s_vert[26], s_vert[3], s_vert[4],
                    s_vert[5], radius);
            fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
            fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
        }     /* if (docell) */
    }         /* for(nocell  */

    if (Display_axes) {

        nvert = 0;                     /* initialize vertex list */
        generate_triple();
        if (radius == 0.) radius = 0.0002f *Scale;
        strcpy(col_cell_v,drvui->col_cell);
        Transform_VRML_Color(col_cell_v);
        strcpy(col_cell_p,drvui->col_cell);
        Transform_POV_Color(col_cell_p);

/* draw vector triple in POV */
        fprintf (drvui->fpoutp, "/* drawing vector triple */");
        fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f \n",s_vert[0], s_vert[1], s_vert[2],
              s_vert[3], s_vert[4],s_vert[5], radius);
        fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
        fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
        fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f \n",s_vert[0], s_vert[1], s_vert[2],
              s_vert[6], s_vert[7],s_vert[8], radius);
        fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
        fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
        fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f \n",s_vert[0], s_vert[1], s_vert[2],
              s_vert[9], s_vert[10],s_vert[11], radius);
        fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
        fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
/* draw vector triple in VRML */
        if (Vrml2) {
            fprintf (drvui->fpoutv, " Shape{\n");
            fprintf (drvui->fpoutv, "  geometry IndexedLineSet { coord Coordinate{ point [\n");
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[0], s_vert[1],s_vert[2]);
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[3], s_vert[4],s_vert[5]);
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[6], s_vert[7],s_vert[8]);
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f]\n  }\n", s_vert[9], s_vert[10], s_vert[11]);
            fprintf (drvui->fpoutv, "  coordIndex [");
            fprintf (drvui->fpoutv, "0,3,0,2,0,1,-1]\n");
            fprintf (drvui->fpoutv, "   color Color { color [%s]}\n   colorIndex[0]\n"
                    "   colorPerVertex FALSE\n",col_cell_v);
            fprintf (drvui->fpoutv, "  }\n");
            fprintf (drvui->fpoutv, " }\n");
        } else {
            fprintf (drvui->fpoutv, " Separator{\n");
            fprintf (drvui->fpoutv, "  Material{\n");
            fprintf (drvui->fpoutv, "   diffuseColor %s\n", col_cell_v);
            fprintf (drvui->fpoutv, "  }\n");
            fprintf (drvui->fpoutv, "  Coordinate3{ point[\n");
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[0], s_vert[1],s_vert[2]);
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[3], s_vert[4],s_vert[5]);
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[6], s_vert[7],s_vert[8]);
            fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f]\n  }\n", s_vert[9], s_vert[10], s_vert[11]);
            fprintf (drvui->fpoutv, "  IndexedLineSet{coordIndex[");
            fprintf (drvui->fpoutv, "0,3,0,2,0,1,-1]\n");
            fprintf (drvui->fpoutv, "  }\n");
            fprintf (drvui->fpoutv, " }\n");
        }
    }

    nvert = 0;                     /* empty vertex list */
}                                /* end of draw_cell */

/* ************************************************************** */
/* ************************************************************** */

void generate_lsq_planes (void)

/* routine to calculate and display a least squares plane through a set of atoms */
{
  int i, j, jj, k, l, m;
  float d;
  float glr,glg,glb;
  char thecolor[40];
  float c[3],s[6],vects[3][3],vals[3];
  double vx[50][15],vy[50][15],vz[50][15];
  double vsx[50][15],vsy[50][15],vsz[50][15];
  double t0[3],t1[3],t2[3],t3[3],p0[3],p1[3],p2[3],p3[3];
  float dummy;
  int nsets=0;
  double sx,sy,sz;
  int nx,add,nn[50];
  int numplanes=0;
  double pn0[MAX_PLANE],pn1[MAX_PLANE],pn2[MAX_PLANE];
  double v1n,v2n,dot,angval;
   
  if (drvui->nbplane==1) return;

  for (m=1;m<drvui->nbplane;m++) {

    nsets = 0;

/* first pass through atom list to find maximum number of planes to expect */
/* (numbers may vary for individual atoms due to view box clipping) */

    for (i = 0; i <= natom; ++i) {

      if (drvui->atom_fn[i] != drvui->frame_no) continue;

      add=0;
      for (j=0;j<drvui->nbatoms[m];j++){ /* only if part of l.s. plane */
        if (  (check_atom_name(drvui->atom_l[i],drvui->bplane_t[m][j])==1)
           && ( drvui->sv_atom_n[i]== drvui->bplane_n[m][j] || drvui->bplane_n[m][j]<0) )
          add=1;
      }
      if (add==0) continue;

      nvert=0;
      find_all_in_box (i); /* find all copies of this atom in the display box */

      if (nvert ==0) {
        Error_Box("At least one atom of the l.s. plane is not in the current display box");
        return;
      }
      if (nvert>nsets) nsets=nvert;
    }

    if (nsets >50) {
        Error_Box("Too many copies (max. 50) of a best fitting plane.");
        return;
    }

/* second pass through master atom list: */
/* assemble atom lists for each copy of the current plane in the view box */

    nx = 0;

    for (i = 0; i <= natom; ++i) {

      if (drvui->atom_fn[i] != drvui->frame_no) continue;

      add=0;
      for (j=0;j<drvui->nbatoms[m];j++){ /* only if part of l.s. plane */
        if (  (check_atom_name(drvui->atom_l[i],drvui->bplane_t[m][j])==1)
           && ( drvui->sv_atom_n[i]== drvui->bplane_n[m][j] || drvui->bplane_n[m][j]<0) )
          add=1;
      }
      if (add==0) continue;

      nvert=0;
      find_all_in_box (i); /* find all copies of this atom in the display box */

      if (nx==0) { /* when lists are empty, add one copy to each */
        for (jj=0;jj<nsets;jj++){
          vx[jj][0]=o_vert[3*jj]-origin[0];
          vy[jj][0]=o_vert[3*jj+1]-origin[1];
          vz[jj][0]=o_vert[3*jj+2]-origin[2];
          vsx[jj][0]=s_vert[3*jj];
          vsy[jj][0]=s_vert[3*jj+1];
          vsz[jj][0]=s_vert[3*jj+2];
          nn[jj]=1;
        }
        nx=1;
      } else {
        for (jj=0;jj<nvert;jj++){      /* find a list that already contains */
          add=0;                       /* atoms within bonding distance */
          for (k=0;k<nsets;k++) {
            for (l=0;l<nx;l++) {
              d=(float)((s_vert[3*jj]-vsx[k][l])*(s_vert[3*jj]-vsx[k][l])
                 +(s_vert[3*jj+1]-vsy[k][l])*(s_vert[3*jj+1]-vsy[k][l])
                 +(s_vert[3*jj+2]-vsz[k][l])*(s_vert[3*jj+2]-vsz[k][l]));
              if (d >0. && d < 1.6*1.6) {
                vx[k][nx]=o_vert[3*jj]-origin[0];
                vy[k][nx]=o_vert[3*jj+1]-origin[1];
                vz[k][nx]=o_vert[3*jj+2]-origin[2];
                vsx[k][nx]=s_vert[3*jj];
                vsy[k][nx]=s_vert[3*jj+1];
                vsz[k][nx]=s_vert[3*jj+2];
                nn[k]++;
                add=1;
                break;
              }
            }
            if (add==1) continue;
          }
        }
        nx++;
        if (nx >14) {
          Error_Box("Cannot handle more than 15 atoms in a best fitting plane.");
          return;
        }
      }
    } /* loop over all atoms i in master list */

/* calculate and render the best fitting plane for each set of atoms */

    for (i=0;i<nsets;i++) {
      if (nn[i]<drvui->nbatoms[m]) continue; /* drop incomplete (clipped) planes */

      c[0]=c[1]=c[2]=0.;
      sx=sy=sz=0.;
      s[0]=s[1]=s[2]=s[3]=s[4]=s[5]=0.;
      for (k=0;k<nx;k++) {
        c[0] += (float)vsx[i][k];
        c[1] += (float)vsy[i][k];
        c[2] += (float)vsz[i][k];
        sx+=vx[i][k];
        sy+=vy[i][k];
        sz+=vz[i][k];
      }
      c[0] /= nx;
      c[1] /= nx;
      c[2] /= nx;
      sx /= nx;
      sy /= nx;
      sz /= nx;
      for (k=0;k<nx;k++) {
        s[0] += (float)((sx-vx[i][k])*(sx-vx[i][k]));
        s[1] += (float)((sy-vy[i][k])*(sy-vy[i][k]));
        s[2] += (float)((sz-vz[i][k])*(sz-vz[i][k]));
        s[3] += (float)((sx-vx[i][k])*(sy-vy[i][k]));
        s[4] += (float)((sx-vx[i][k])*(sz-vz[i][k]));
        s[5] += (float)((sy-vy[i][k])*(sz-vz[i][k]));
      }

      eigen (&dummy,s,vals,vects);

      d = (float)sqrt( vects[0][0]*vects[0][0]+vects[0][1]*vects[0][1]+vects[0][2]*vects[0][2]);
      vects[0][0] /= d;
      vects[0][1] /= d;
      vects[0][2] /= d;
      d = (float)sqrt( vects[1][0]*vects[1][0]+vects[1][1]*vects[1][1]+vects[1][2]*vects[1][2]);
      vects[1][0] /= d;
      vects[1][1] /= d;
      vects[1][2] /= d;
      d = (float)sqrt( vects[2][0]*vects[2][0]+vects[2][1]*vects[2][1]+vects[2][2]*vects[2][2]);
      vects[2][0] /= d;
      vects[2][1] /= d;
      vects[2][2] /= d;

      fprintf(drvui->flout,"least squares plane %2d(%2d): center at    %f %f %f\n",m,i+1,c[0],c[1],c[2]);
      fprintf(drvui->flout,"                            eigenvalues  %f %f %f\n",vals[0],vals[1],vals[2]);
      fprintf(drvui->flout,"                            plane normal %f %f %f\n",vects[0][0],vects[0][1],vects[0][2]);
      fprintf(drvui->flout,"                            1st vector   %f %f %f\n",vects[1][0],vects[1][1],vects[1][2]);
      fprintf(drvui->flout,"                            2nd vector   %f %f %f\n",vects[2][0],vects[2][1],vects[2][2]);

      /* apply scaling factors to generate initial cornerpoints of polygon */
      t0[0]=c[0]+drvui->bplane_d1[m]*vects[2][0];
      t0[1]=c[1]+drvui->bplane_d1[m]*vects[2][1];
      t0[2]=c[2]+drvui->bplane_d1[m]*vects[2][2];
      t1[0]=c[0]+drvui->bplane_d2[m]*vects[1][0];
      t1[1]=c[1]+drvui->bplane_d2[m]*vects[1][1];
      t1[2]=c[2]+drvui->bplane_d2[m]*vects[1][2];
      t2[0]=c[0]-drvui->bplane_d1[m]*vects[2][0];
      t2[1]=c[1]-drvui->bplane_d1[m]*vects[2][1];
      t2[2]=c[2]-drvui->bplane_d1[m]*vects[2][2];
      t3[0]=c[0]-drvui->bplane_d2[m]*vects[1][0];
      t3[1]=c[1]-drvui->bplane_d2[m]*vects[1][1];
      t3[2]=c[2]-drvui->bplane_d2[m]*vects[1][2];
      /* treat these as midpoints on the edges of the final rectangle */
      /* to make width and height factors behave as expected */
      for (jj=0;jj<3;jj++) {
        p0[jj]=c[jj]+(t0[jj]-t1[jj])/2.;
        p1[jj]=c[jj]+(t1[jj]-t2[jj])/2.;
        p2[jj]=c[jj]+(t2[jj]-t3[jj])/2.;
        p3[jj]=c[jj]+(t3[jj]-t0[jj])/2.;
      }

      strncpy(thecolor,drvui->bplane_col[m],39);
      Transform_VRML_Color(thecolor);
      sscanf(thecolor,"%f %f %f",&glr,&glg,&glb);

#ifdef OPENGL
      if (gl_view) {
        glPushMatrix();
        glDisable(GL_LIGHTING);
        glColor3f (glr, glg, glb);
        glBegin(GL_QUADS);
        glNormal3f(c[0]+vects[0][0],c[1]+vects[0][1],c[2]+vects[0][2]);
        glVertex3f((float)p0[0],(float)p0[1],(float)p0[2]);
        glVertex3f((float)p1[0],(float)p1[1],(float)p1[2]);
        glVertex3f((float)p2[0],(float)p2[1],(float)p2[2]);
        glVertex3f((float)p3[0],(float)p3[1],(float)p3[2]);
        glEnd();
        glEnable(GL_LIGHTING);
        glPopMatrix();
      }
#endif

      fprintf(drvui->fpoutp,"/* L.S. Plane */\n");
      fprintf(drvui->fpoutp,"triangle {< %8.5f, %8.5f, %8.5f>,\n",
        c[0],c[1],c[2]);
      fprintf(drvui->fpoutp,"<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>\n",
        p0[0],p0[1],p0[2],p1[0],p1[1],p1[2]);
      fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n",drvui->bplane_col[m]);
      fprintf(drvui->fpoutp," }\n");
      fprintf(drvui->fpoutp,"triangle {< %8.5f, %8.5f, %8.5f>,\n",
        c[0],c[1],c[2]);
      fprintf(drvui->fpoutp,"<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>\n",
        p1[0],p1[1],p1[2],p2[0],p2[1],p2[2]);
      fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n",drvui->bplane_col[m]);
      fprintf(drvui->fpoutp," }\n");
      fprintf(drvui->fpoutp,"triangle {< %8.5f, %8.5f, %8.5f>,\n",
        c[0],c[1],c[2]);
      fprintf(drvui->fpoutp,"<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>\n",
        p2[0],p2[1],p2[2],p3[0],p3[1],p3[2]);
      fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n",drvui->bplane_col[m]);
      fprintf(drvui->fpoutp," }\n");
      fprintf(drvui->fpoutp,"triangle {< %8.5f, %8.5f, %8.5f>,\n",
        c[0],c[1],c[2]);
      fprintf(drvui->fpoutp,"<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>\n",
        p3[0],p3[1],p3[2],p0[0],p0[1],p0[2]);
      fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n",drvui->bplane_col[m]);
      fprintf(drvui->fpoutp," }\n");

      fprintf(drvui->fpoutv,"# L.S. Plane\n");
      if (Vrml2) {
        fprintf (drvui->fpoutv, " Shape {");
        fprintf (drvui->fpoutv, "appearance Appearance {\n");
        fprintf (drvui->fpoutv,"   material Material {diffuseColor %s} \n \n",
          thecolor);
      } else {
        fprintf (drvui->fpoutv, " Separator {\n        ");
        fprintf (drvui->fpoutv, "  Material {\n");
        fprintf (drvui->fpoutv, "   diffuseColor %s \n  }\n",thecolor);
      }
      if (Vrml2)
        fprintf (drvui->fpoutv,"}\n geometry IndexedFaceSet { coord Coordinate{ point [\n");
      else
        fprintf (drvui->fpoutv, "Coordinate3 { point [\n");

      fprintf (drvui->fpoutv, " %5.3f %5.3f %5.3f,\n",
        p0[0],p0[1],p0[2]);
      fprintf (drvui->fpoutv, " %5.3f %5.3f %5.3f,\n",
        p1[0],p1[1],p1[2]);
      fprintf (drvui->fpoutv, " %5.3f %5.3f %5.3f,\n",
        p2[0],p2[1],p2[2]);
      fprintf (drvui->fpoutv, " %5.3f %5.3f %5.3f]\n",
        p3[0],p3[1],p3[2]);

      if (Vrml2) {
        fprintf (drvui->fpoutv, " }\n");
        fprintf (drvui->fpoutv, " coordIndex [0,1,2,3,-1]\n");
        fprintf (drvui->fpoutv, " solid FALSE\n");
        fprintf (drvui->fpoutv, " convex TRUE\n");
        fprintf (drvui->fpoutv, " creaseAngle 1.5708\n");
        fprintf (drvui->fpoutv, "  }\n }\n");
      } else {
        fprintf (drvui->fpoutv, "}\n IndexedFaceSet { coordIndex [0,1,2,3,-1]\n");
        fprintf (drvui->fpoutv,"  }\n }\n");
      }
      if (i==0) {
        pn0[m]=vects[0][0];
        pn1[m]=vects[0][1];
        pn2[m]=vects[0][2];
      }
      numplanes ++;

    } /* for all l.s. planes */

  } /* for all bestplane keywords */

  fprintf(drvui->flout,"Angles between planes:\n");
  for (i=1;i<m;i++)  {
    v1n=pn0[i]*pn0[i]+pn1[i]*pn1[i]+pn2[i]*pn2[i];
    v1n=sqrt(v1n);
    for (j=i+1;j<m;j++) {
      v2n=pn0[j]*pn0[j]+pn1[j]*pn1[j]+pn2[j]*pn2[j];
      v2n=sqrt(v2n);
      dot = pn0[i]*pn0[j]+pn1[i]*pn1[j]+pn2[i]*pn2[j];
      angval= RAD * acos(dot /(v1n*v2n));
      fprintf(drvui->flout,"Plane %d - Plane %d: %5.3f\n",i,j,angval);
    }
  }

  fprintf(drvui->flout,"Generated %d best fitting planes.\n",numplanes);
}

#ifdef OPENGL
/* ************************************************************** */
/* ************************************************************** */

void draw_GL_triple(void)
{
/* draw the lines for the vector triple in openGL */
    int i;
    float at[3], df[3];
    float radius = rad_cell * 0.01f *Scale;
    char col_cell_v[40];

    strcpy(col_cell_v,drvui->col_cell);
    Transform_VRML_Color(col_cell_v);
    
    for (i=0; i<3; i++) {
        at[i] = s_vert[i];
        df[i] = s_vert[3 + i] - at[i];
    }
    push_cylinder(df, at, radius, col_cell_v);
    for (i=0; i<3; i++) {
        at[i] = s_vert[i];
        df[i] = s_vert[6 + i] - at[i];
    }
    push_cylinder(df, at, radius, col_cell_v);
    for (i=0; i<3; i++) {
        at[i] = s_vert[i];
        df[i] = s_vert[9 + i] - at[i];
    }
    push_cylinder(df, at, radius, col_cell_v);
}
#endif

/* ************************************************************** */
/* ************************************************************** */

void generate_spheres (void)

/* routine to generate spherical objects and add then to edit list */
{
  float *radii;					 /* dynamic array for sphere radii */
  int Sphere_Count;				 /* Counter for number output */
  int i, j, k;					 /* loop counters */

  Sphere_Count = 0;
  for (i = 1; i < drvui->nsphere; ++i) {  /* loop through spheres */
    if (drvui->sphere_fn[i] != drvui->frame_no) continue;  /* skip if not in this frame */
    nvert = 0;                /* clear the vertex list */
    for (j = 0; j < natom; ++j)    /* loop through atoms */
      if ((drvui->atom_n[j] & 255) == i)
        find_all_in_box (j);
    if (nvert > 0) {          /* if spheres present */
/* allocate storage for radii of spheres */
      radii = zalloc ((unsigned) (nvert * sizeof (float)));
      for (k = 0; k < nvert; ++k)  /* copy radii info to storage */
        radii[k] = drvui->sphere_size[i] * Sphere_Mult;
      Output_Spheres (radii, i);
      free (radii);
    }
    Sphere_Count += nvert;
  }
  if (Sphere_Count > 0) {
    printf ("%4d spheres.\n", Sphere_Count);
    fprintf (drvui->flout, "Generated %4d spheres.\n", Sphere_Count);
  }
}						 /* end of generate_spheres */

/* ************************************************************** */
/* ************************************************************** */

void generate_triple(void)
{
/* generate the vertices for the vector triple */
    
    int i, j;
    float vert[3];
   
    nvert = 0;
    for (i = 0; i < 3; i++)
      vert[i] = 0.0f;
    add_vert_nc (vert);                 /* place origin at 0,0,0 in list */
    vert[2] = 1.0f / drvui->lat_con[2];
    add_vert_nc (vert);                 /* place 1 A vector along 001 */
    vert[2] = 0.0f;
    vert[1] = 1.0f / drvui->lat_con[1];
    add_vert_nc (vert);                 /* place 1 A vector along 010 */
    vert[1] = 0.0f;
    vert[0] = 1.0f / drvui->lat_con[0];
    add_vert_nc (vert);                 /* place 1 A vector along 100 */
    vert[2] = 1.4f / drvui->lat_con[2];
    vert[0] = 0.0f;
    add_vert_nc (vert);                 /* place 1.4 A vector along 001 */
    vert[2] = 0.0f;
    vert[1] = 1.4f / drvui->lat_con[1];
    add_vert_nc (vert);                 /* place 1.4 A vector along 010 */
    vert[1] = 0.0f;
    vert[0] = 1.4f / drvui->lat_con[0];
    add_vert_nc (vert);                 /* place 1.4 A vector along 100 */
    for (i=nvert-1; i>=0; --i) {             /* Translate coordinates */
      for (j = 0; j < 3; j++) {
        s_vert[3 * i + j] += offset[j] - s_vert[j];
      }
    }
}

/* ************************************************************** */
/* ************************************************************** */

void import_schakal (char input[100], int in_line, int Quick, int frame_no)

/* routine to extract structural information from a SCHAKAL data file */
{
  FILE *impin;
  char string[100];				 /* temporary string */
  char filename[100];				 /* string for filename */
  int intype;					 /* type of datum read */
  int i, j;					 /* temporary */

  if (in_line) {
    impin = drvui->fpin;
  }
  else {
    strcpy (filename,input);
    if ((impin = fopen (filename, "r")) == NULL) {
      fprintf (stderr, "Cannot Open SCHAKAL Import File, Run aborted.\n");
      delay (5);
      exit (0);
    }
  }
  if (!Quick)
    fprintf (drvui->flout, "****** Now interpreting SCHAKAL-style data\n");
  while (!feof (impin)) {
    if (!fgets (input, 100, impin)) {		 /* read a line */
      fprintf (stderr, "Error reading file.");
      delay (5);
      exit (1);
    }
    intype = 0;
    for (i = 0; i < 4; i++) {
      if (input[i] == '\0')
	intype = 3;
    }
    if (strncmp (input, "TITL", 4) == 0)
      intype = 1;				 /* echo and ignore */
    if (strncmp (input, "titl ", 4) == 0)
      intype = 1;				 /* echo and ignore */
    if (strncmp (input, "CELL", 4) == 0)
      intype = 2;
    if (strncmp (input, "cell", 4) == 0)
      intype = 2;
    if (strncmp (input, "ATOM", 4) == 0)
      intype = 3;				 /* atom coordinate card */
    if (strncmp (input, "atom", 4) == 0)
      intype = 3;
    if (strncmp (input, "END", 3) == 0)
      intype = 4;				 /* trigger exit */
    if (strncmp (input, "end", 3) == 0)
      intype = 4;				 /* trigger exit */


    switch (intype) {
    case 1:
      for (i = 0; i < 51; i++)
	if (input[i] == '\n')
	  input[i] = ' ';
      input[50] = '\0';				 /* truncate title */
      if (!Quick)
	fprintf (drvui->flout, "******      %s\n", input);
      break;					 /* ignore title except for listing */
    case 2:
      for (i = 3; i <= 5; ++i)
        drvui->lat_con[i] = (float) 0.;
      (void) sscanf (input, "%s %f %f %f %f %f %f", string, &drvui->lat_con[0],
		     &drvui->lat_con[1], &drvui->lat_con[2], &drvui->lat_con[3],
             &drvui->lat_con[4], &drvui->lat_con[5]);
      if (!Quick)
	fprintf (drvui->flout,
		 "******      Cell     %8.5f %8.5f %8.5f %8.3f %8.3f %8.3f\n",
		 drvui->lat_con[0], drvui->lat_con[1], drvui->lat_con[2], drvui->lat_con[3],
		 drvui->lat_con[4], drvui->lat_con[5]);
      break;
    case 3:					 /* read an atom line */
      (void) sscanf (input, "%s %c%c%c%c %f %f %f", string,
		     &drvui->atom_l[natom][0], &drvui->atom_l[natom][1],
		     &drvui->atom_l[natom][2], &drvui->atom_l[natom][3],
		     &drvui->atom_xyz[natom][0], &drvui->atom_xyz[natom][1],
		     &drvui->atom_xyz[natom][2]);
      drvui->atom_fn[natom] = frame_no;
      drvui->atom_n[natom] = 0;
      for (j = 0; j < 4; j++)			 /* extract atom number from SCHAKAL name */
	if (drvui->atom_l[natom][j] >= '0' && drvui->atom_l[natom][j] <= '9') {
	  drvui->atom_n[natom] = drvui->atom_n[natom] * 10 + (int) drvui->atom_l[natom][j] - 48;
	  drvui->atom_l[natom][j] = ' ';
	}
      if (drvui->atom_n[natom] == 0)
        drvui->atom_n[natom] = 1;
      if (!Quick)
        fprintf (drvui->flout,
		   "******      Atom %c%c%c%c%3d %8.5f %8.5f %8.5f\n",
		   drvui->atom_l[natom][0], drvui->atom_l[natom][1],
           drvui->atom_l[natom][2], drvui->atom_l[natom][3],
           drvui->atom_n[natom], drvui->atom_xyz[natom][0],
		   drvui->atom_xyz[natom][1], drvui->atom_xyz[natom][2]);
      natom++;
      if (natom >= MAX_ATOM) {
        fprintf (stderr,"Too many atoms in asymmetric unit - increase MAX_ATOM.\n");
        delay (5);
        exit (0);
      }
      break;
    case 4:
      if (!Quick)
	fprintf (drvui->flout, "****** END - end of SCHAKAL dataset\n");
      if (!in_line)
	(void) fclose (impin);
      return;
    default:
      break;					 /* end of default */
    }						 /* end of switch */
  }
}						 /* end of import_schakal */

/* ************************************************************** */
/* ************************************************************** */

void Transform_POV_Color(char * color)
{
/* Transform a POV color is specified in the form f1 f2 f2 {filter f} format,
 * where {filter} is optional. If a color name is specified, return with no
 * action.                                                                   */
    float red, green, blue, filter;
    char working[120], string[20];
    char *p;

    if (!strtod(color,&p) && p == color)
        return;                          /* we have the color name form */

    strcpy(working, color);              /* make a working copy */
    (void)sscanf(working, "%f %f %f", &red, &green, &blue);

    if ((p = strstr(working,"filter"))) { /* if "filter" given, return rgbf<...> form */
        (void)sscanf(p, "%s %f", string,&filter);
        sprintf(working,"rgbf <%.4f,%.4f,%.4f,%.4f>",red,green,blue,filter);
    } else {                             /* return rgb <...> form */
        sprintf(working,"rgb <%.4f,%.4f,%.4f>",red,green,blue);
    }
    strcpy(color, working);              /* reformatted color to caller */
}
    
/* ************************************************************** */
/* ************************************************************** */

void Transform_VRML_Color (char *input)
{
  char *transp_start;
  char transp[20];
  float fval;
  int i = strlen(input);
  char *a;
  float thered,thegreen,theblue;

/*
   VRML routine to transform VRML (and openGL) colors into RGB triples
   derived from 'colors.inc' of POV

   If the color information is "float1 float2 float3", then extract the rgb
   values from the input.   LWF 9/30/05
*/

  if (input[i-1] < ' ') input[i-1] = 0;        /* get rid of any ^J character */
  i = strlen(input);
  if (input[i-1] < ' ') input[i-1] = 0;        /* get rid of any ^J character */
  transp[0] = '\0';
  transp_start = strstr (input, "filter");
  if (transp_start) {
    fval = (float) atof (transp_start + 6);
    sprintf (transp, " transparency %5.2f", fval / 2.);
  }
  if (strtod(input,&a) || a != input) {             /* rgb form of color */
      sscanf(input,"%f %f %f", &thered, &thegreen, &theblue);
      sprintf(input,"%.6f %.6f %.6f",thered,thegreen,theblue);
      strcat(input, transp);
      return;
  }
  if (strncmp (input, "LightGray", 9) == 0) {
    strncpy (input, ".658824 .658824 .658824\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "LightGrey", 9) == 0) {
    strncpy (input, ".658824 .658824 .658824\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "VLightGrey", 10) == 0) {
    strncpy (input, ".8 .8 .8\0", 9);
    strcat (input, transp);
    return;
  }

  if (strncmp (input, "White", 5) == 0) {
    strncpy (input, "1 1 1\0", 6);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Red", 3) == 0) {
    strncpy (input, "1 0 0\0", 6);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "GreenYellow", 11) == 0) {
    strncpy (input, ".576471 .858824 .439216\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Green", 5) == 0) {
    strncpy (input, "0 1 0\0", 6);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "BlueViolet", 10) == 0) {
    strncpy (input, ".62352 .372549 .623529\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Blue", 3) == 0) {
    strncpy (input, "0 0 1\0", 6);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "YellowGreen", 11) == 0) {
    strncpy (input, ".6 .8 .196078\0", 14);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Yellow", 6) == 0) {
    strncpy (input, "1 1 0\0", 6);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Cyan", 3) == 0) {
    strncpy (input, "0 1 1\0", 6);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Magenta", 7) == 0) {
    strncpy (input, "1 0 1\0", 6);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Black", 5) == 0) {
    strncpy (input, "0 0 0\0", 6);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "OrangeRed", 9) == 0) {
    strncpy (input, "1.0 .498039 0.\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Orange", 6) == 0) {
    strncpy (input, "1 .5 .0\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Brown", 5) == 0) {
    strncpy (input, ".647059 .164706 .164706\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray05", 6) == 0) {
    strncpy (input, ".05 .05 .05\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray10", 6) == 0) {
    strncpy (input, ".10 .10 .10\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray15", 6) == 0) {
    strncpy (input, ".15 .15 .15\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray20", 6) == 0) {
    strncpy (input, ".20 .20 .20\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray25", 6) == 0) {
    strncpy (input, ".25 .25 .25\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray30", 6) == 0) {
    strncpy (input, ".30 .30 .30\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray35", 6) == 0) {
    strncpy (input, ".35 .35 .35\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray40", 6) == 0) {
    strncpy (input, ".40 .40 .40\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray45", 6) == 0) {
    strncpy (input, ".45 .45 .45\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray50", 6) == 0) {
    strncpy (input, ".50 .50 .50\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray55", 6) == 0) {
    strncpy (input, ".55 .55 .55\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray60", 6) == 0) {
    strncpy (input, ".60 .60 .60\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray65", 6) == 0) {
    strncpy (input, ".65 .65 .65\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray70", 6) == 0) {
    strncpy (input, ".70 .70 .70\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray75", 6) == 0) {
    strncpy (input, ".75 .75 .75\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray80", 6) == 0) {
    strncpy (input, ".80 .80 .80\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray85", 6) == 0) {
    strncpy (input, ".85 .85 .85\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray90", 6) == 0) {
    strncpy (input, ".90 .90 .90\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray95", 6) == 0) {
    strncpy (input, ".95 .95 .95\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray", 4) == 0) {
    strncpy (input, ".752941 .752941 .752941\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Grey", 4) == 0) {
    strncpy (input, ".752941 .752941 .752941\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "DimGray", 7) == 0) {
    strncpy (input, ".329412 .329412 .329412\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "DimGrey", 7) == 0) {
    strncpy (input, ".329412 .329412 .329412\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Aquamarine", 10) == 0) {
    strncpy (input, ".439216 .858824 .576471\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "CadetBlue", 9) == 0) {
    strncpy (input, ".372549 .623529 .623529\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Coral", 5) == 0) {
    strncpy (input, "1.0 .498039 .0\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "CornflowerBlue", 14) == 0) {
    strncpy (input, ".258824 .258824 .435294\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "DarkGreen", 9) == 0) {
    strncpy (input, ".184314 .309804 .184314\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "DarkOliveGreen", 14) == 0) {
    strncpy (input, ".309804 .309804 .184314\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "DarkOrchid", 10) == 0) {
    strncpy (input, ".6 .196078 .8\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "DarkSlateBlue", 13) == 0) {
    strncpy (input, ".119608 .137255 .556863\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "DarkSlateGray", 13) == 0) {
    strncpy (input, ".184314 .309804 .309804\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "DarkSlateGrey", 13) == 0) {
    strncpy (input, ".184314 .309804 .309804\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "DarkTurquoise", 13) == 0) {
    strncpy (input, ".439216 .576471 .858824\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Firebrick", 9) == 0) {
    strncpy (input, ".556863 .137255 .137255\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "ForestGreen", 11) == 0) {
    strncpy (input, ".137255 .556863 .137255\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Goldenrod", 9) == 0) {
    strncpy (input, ".858824 .858824 .439216\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gold", 4) == 0) {
    strncpy (input, ".8 .498039 .196078\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "IndianRed", 9) == 0) {
    strncpy (input, ".309804 .184314 .184314\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Khaki", 5) == 0) {
    strncpy (input, ".623529 .623529 .372549\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "LightBlue", 9) == 0) {
    strncpy (input, ".74902 .847059 .847059\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "LightSteelBlue", 14) == 0) {
    strncpy (input, ".560784 .560784 .737255\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "LimeGreen", 9) == 0) {
    strncpy (input, ".196078 .8 .196078\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Maroon", 6) == 0) {
    strncpy (input, ".556863 .137255 .419608\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "MediumAquamarine", 16) == 0) {
    strncpy (input, ".196078 .8 .6\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "MediumBlue", 10) == 0) {
    strncpy (input, ".196078 .196078 .8\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "MediumForestGreen", 17) == 0) {
    strncpy (input, ".419608 .556863 .137255\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "MediumGoldenrod", 15) == 0) {
    strncpy (input, ".917647 .917647 .678431\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "MediumOrchid", 12) == 0) {
    strncpy (input, ".576471 .439216 .858824\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "MediumSeaGreen", 14) == 0) {
    strncpy (input, ".258824 .435294 .258824\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "MediumSlateBlue", 15) == 0) {
    strncpy (input, ".498039 0.0 1.0\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "MediumSpringGreen", 17) == 0) {
    strncpy (input, ".498039 1.0 0\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "MediumTurquoise", 15) == 0) {
    strncpy (input, ".439216 .858824 .858824\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "MediumVioletRed", 15) == 0) {
    strncpy (input, ".858824 .439216 .576471\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "MidnightBlue", 12) == 0) {
    strncpy (input, ".184314 .184314 .309804\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "NavyBlue", 8) == 0) {
    strncpy (input, ".137255 .137255 .556863\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Navy", 4) == 0) {
    strncpy (input, ".137255 .137255 .556863\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Orchid", 6) == 0) {
    strncpy (input, ".858824 .439216 .858824\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "PaleGreen", 9) == 0) {
    strncpy (input, ".560784 .737255 .560784\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Pink", 4) == 0) {
    strncpy (input, ".737255 .560784 .560784\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Plum", 4) == 0) {
    strncpy (input, ".917647 .678431 .917647\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Salmon", 6) == 0) {
    strncpy (input, ".435294 .258824 .258824\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "SeaGreen", 8) == 0) {
    strncpy (input, ".137255 .556863 .419608\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Sienna", 6) == 0) {
    strncpy (input, ".556863 .419608 .137255\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "SkyBlue", 7) == 0) {
    strncpy (input, ".196078 .6 .8\0", 14);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "SlateBlue", 9) == 0) {
    strncpy (input, "0 .498039 1\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "SpringGreen", 11) == 0) {
    strncpy (input, "0 1 .498039\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "SteelBlue", 9) == 0) {
    strncpy (input, ".137255 .419608 .556863\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Tan", 3) == 0) {
    strncpy (input, ".858824 .576471 .439216\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Thistle", 7) == 0) {
    strncpy (input, ".847059 .74902 .847059\0", 23);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Turquoise", 9) == 0) {
    strncpy (input, ".678431 .917647 .917647\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "VioletRed", 9) == 0) {
    strncpy (input, ".8 .196078 .6\0", 14);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Violet", 6) == 0) {
    strncpy (input, ".309804 .184314 .309804\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Wheat", 5) == 0) {
    strncpy (input, ".847059 .847059 .74902\0", 23);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "SummerSky", 9) == 0) {
    strncpy (input, ".22 .69 .87\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "RichBlue", 8) == 0) {
    strncpy (input, ".35 .35 .67\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Brass", 5) == 0) {
    strncpy (input, ".71 .65 .26\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Copper", 6) == 0) {
    strncpy (input, ".72 .45 .20\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Bronze2", 7) == 0) {
    strncpy (input, ".65 .49 .24\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Bronze", 6) == 0) {
    strncpy (input, ".55 .47 .14\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Silver", 6) == 0) {
    strncpy (input, ".90 .91 .98\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "BrightGold", 10) == 0) {
    strncpy (input, ".85 .85 .10\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "OldGold", 7) == 0) {
    strncpy (input, ".81 .71 .23\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Feldspar", 8) == 0) {
    strncpy (input, ".82 .57 .46\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Quartz", 6) == 0) {
    strncpy (input, ".85 .85 .95\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Mica", 4) == 0) {
    strncpy (input, "0 0 0\0", 6);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "NeonPink", 8) == 0) {
    strncpy (input, "1 .43 .78\0", 10);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "DarkPurple", 10) == 0) {
    strncpy (input, ".53 .12 .47\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "NeonBlue", 8) == 0) {
    strncpy (input, ".3 .3 1\0", 8);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "CoolCopper", 10) == 0) {
    strncpy (input, ".85 .53 .10\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "MandarinOrange", 14) == 0) {
    strncpy (input, ".89 .47 .20\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "LightWood", 9) == 0) {
    strncpy (input, ".91 .76 .65\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "MediumWood", 10) == 0) {
    strncpy (input, ".65 .50 .39\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "DarkWood", 8) == 0) {
    strncpy (input, ".52 .37 .26\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "SpicyPink", 9) == 0) {
    strncpy (input, "1 .11 .68\0", 10);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "SemiSweetChoc", 13) == 0) {
    strncpy (input, ".42 .26 .15\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "BakersChoc", 10) == 0) {
    strncpy (input, ".36 .20 .09\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Flesh", 5) == 0) {
    strncpy (input, ".96 .80 .69\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "NewTan", 6) == 0) {
    strncpy (input, ".92 .78 .62\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "NewMidnightBlue", 15) == 0) {
    strncpy (input, "0 0 .61\0", 8);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "VeryDarkBrown", 13) == 0) {
    strncpy (input, ".35 .16 .14\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "DarkBrown", 9) == 0) {
    strncpy (input, ".36 .25 .20\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "DarkTan", 7) == 0) {
    strncpy (input, ".59 .41 .31\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "GreenCopper", 11) == 0) {
    strncpy (input, ".32 .49 .46\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "DkGreenCopper", 13) == 0) {
    strncpy (input, ".29 .46 .43\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "DustyRose", 9) == 0) {
    strncpy (input, ".52 .39 .39\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "HuntersGreen", 12) == 0) {
    strncpy (input, ".13 .37 .31\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Scarlet", 7) == 0) {
    strncpy (input, ".55 .09 .09\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Med_Purple", 10) == 0) {
    strncpy (input, ".73 .16 .96\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Light_Purple", 12) == 0) {
    strncpy (input, ".87 .58 .98\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Very_Light_Purple", 17) == 0) {
    strncpy (input, ".94 .81 .99\0", 12);
    strcat (input, transp);
    return;
  }
  if (Color_Warning <= 5) {
    fprintf (drvui->flout,
	     " **** Warning - Non-standard color \"%s\" encountered. ****\n",
	     input);
    Color_Warning++;
  }
}

/* ************************************************************** */
/* ************************************************************** */

void trim_string(char string[], int len)
{
/* trim string - remove trailing ^J (Windows) and spaces. 'len' is maximum length */
    int i;
    string[len-1] = 0;
    if (strlen(string) == 0) {
        strcpy(string,"White");
        return;
    }
    for (i=0;i<(int)strlen(string);i++) {
        if (string[i] < ' ') {
            string[i] = 0;
            break;
        }
    }
    for (i=strlen(string);i > 0; --i) {
        if (string[i-1] <= ' ') {
            string[i-1] = 0;
        } else
            break;
    }
}

